@@ -19,17 +19,19 @@ class Item(TraceableBaseNode):
             collection (TraceableCollection): Collection for which to generate the nodes.
         """
         self._item = collection.get_item(self['id'])
-        item_id = self._item.get_id()
+        item_id = self._item.identifier
         top_node = self.create_top_node(item_id, app=app)
         dl_node = nodes.definition_list()
         if app.config.traceability_render_attributes_per_item:
             self._process_attributes(dl_node, app)
         if app.config.traceability_render_relationship_per_item:
-            self._process_relationships(collection, dl_node, app)
+            self._process_relationships(dl_node, app)
         if dl_node.children:
             top_node.append(dl_node)
         # Note: content should be displayed during read of RST file, as it contains other RST objects
         self.replace_self(top_node)
+        if app.config.traceability_inspect_item:
+            app.config.traceability_inspect_item(item_id, collection)
 
     def _process_attributes(self, dl_node, app):
         """ Processes all attributes for the given item and adds the list of attributes to the given definition list.
@@ -54,23 +56,20 @@ class Item(TraceableBaseNode):
                 li_node.append(dd_node)
             dl_node.append(li_node)
 
-    def _process_relationships(self, collection, *args):
-        """ Processes all relationships of the item. All targets get listed per relationship.
+    def _process_relationships(self, *args):
+        """ Processes all relationships of the item in natural order.
 
-        Args:
-            collection (TraceableCollection): Collection of all TraceableItems.
+        All targets get naturally sorted per relationship.
         """
-        for rel in collection.iter_relations():
-            targets = self._item.iter_targets(rel)
-            if targets:
-                self._list_targets_for_relation(rel, targets, *args)
+        for rel, targets in self._item.all_relations:
+            self._list_targets_for_relation(rel, targets, *args)
 
     def _list_targets_for_relation(self, relation, targets, dl_node, app):
         """ Add a list with all targets for a specific relation to the given definition list.
 
         Args:
             relation (str): Name of the relation.
-            targets (list): Naturally sorted list of targets to other traceable item(s).
+            targets (iterable): Naturally sorted iterable of targets to other traceable item(s).
             dl_node (nodes.definition_list): Definition list of the item.
             app: Sphinx's application object to use.
         """
@@ -149,22 +148,18 @@ class ItemDirective(TraceableBaseDirective):
         item_node['classes'].append('collapsible_links')  # traceability.js adds the arrowhead button
         if app.config.traceability_collapse_links:
             item_node['classes'].append('collapse')
-
-        target_node = self._store_item_info(target_id, env)
+        if 'class' in self.options:
+            item_node['classes'].extend(self.options.get('class'))
+        item = self._store_item_info(target_id, env)
 
         # Custom callback for modifying items
         if app.config.traceability_callback_per_item:
             app.config.traceability_callback_per_item(target_id, env.traceability_collection)
-
-        # Output content of item to document
-        template = []
-        for line in self.content:
-            template.append('    ' + line)
-        self.state_machine.insert_input(template, self.state_machine.document.attributes['source'])
+        item.clear_state()  # avoid access to the state machine after this point
 
         self.check_caption_flags(item_node, app.config.traceability_item_no_captions)
 
-        return [target_node, item_node]
+        return [item.node, item_node, item.content_node]
 
     def _store_item_info(self, target_id, env):
         """ Stores item info and adds TraceableItem to the collection.
@@ -174,14 +169,14 @@ class ItemDirective(TraceableBaseDirective):
             env (sphinx.environment.BuildEnvironment): Sphinx's build environment.
 
         Returns:
-            (nodes.target) Node object which is bound to the traceable item.
+            TraceableItem: Instantiated TraceableItem.
         """
         target_node = nodes.target('', '', ids=[target_id])
-        item = TraceableItem(target_id)
-        item.set_document(env.docname, self.lineno)
-        item.bind_node(target_node)
-        item.set_caption(self.get_caption())
-        item.set_content('\n'.join(self.content))
+        item = TraceableItem(target_id, state=self.state)
+        item.set_location(env.docname, self.lineno)
+        item.node = target_node
+        item.caption = self.caption
+        item.content = '\n'.join(self.content)
         try:
             env.traceability_collection.add_item(item)
         except TraceabilityException as err:
@@ -196,7 +191,7 @@ class ItemDirective(TraceableBaseDirective):
             related_ids = self.options[rel].split()
             self._add_relation_to_ids(rel, target_id, related_ids, env)
 
-        return target_node
+        return item
 
     def _add_relation_to_ids(self, relation, source_id, related_ids, env):
         """ Adds the given relation between the source id and all related ids.
