@@ -1,16 +1,16 @@
 from docutils.parsers.rst import directives
 
-from mlx.traceability import report_warning
-from mlx.traceability_exception import TraceabilityException
+from mlx.traceability_exception import report_warning, TraceabilityException
 from mlx.traceable_base_directive import TraceableBaseDirective
 from mlx.traceable_base_node import TraceableBaseNode
 
 
 class ItemLink(TraceableBaseNode):
     '''Linking of documentation items'''
+    order = 1  # before ItemRelink
 
     def perform_replacement(self, app, collection):
-        """ Processes the item-link items. The ItemLink node has no final representation, so is removed from the tree.
+        """ The ItemLink node has no final representation, so is removed from the tree.
 
         Args:
             app: Sphinx application object to use.
@@ -18,6 +18,30 @@ class ItemLink(TraceableBaseNode):
         """
         self.replace_self([])
 
+    def apply_effect(self, collection):
+        """ Processes the item-link items, which shall be done before converting anything to docutils and before any
+        item-relink items have been processed.
+
+        Args:
+            collection (TraceableCollection): Collection for which to generate the nodes.
+        """
+        if self['sources']:
+            source_items = self['sources']
+        else:
+            source_items = collection.get_items(self['source'], sort=False)
+        if self['targets']:
+            target_items = self['targets']
+        else:
+            target_items = collection.get_items(self['target'], sort=False)
+        # Processing of the item-link items. They get added as additional relationships
+        # to the existing items. Should be done before converting anything to docutils.
+        for source in source_items:
+            for target in target_items:
+                try:
+                    collection.add_relation(source, self['type'], target)
+                except TraceabilityException as err:
+                    report_warning(err, self['document'], self['line'])
+
 
 class ItemLinkDirective(TraceableBaseDirective):
     """
@@ -26,14 +50,17 @@ class ItemLinkDirective(TraceableBaseDirective):
     Syntax::
 
       .. item-link::
+         :source: regexp
          :sources: list_of_items
+         :target: regexp
          :targets: list_of_items
          :type: relationship_type
-
     """
     # Options
     option_spec = {
+        'source': directives.unchanged,
         'sources': directives.unchanged,
+        'target': directives.unchanged,
         'targets': directives.unchanged,
         'type': directives.unchanged,
     }
@@ -51,23 +78,26 @@ class ItemLinkDirective(TraceableBaseDirective):
         process_options_success = self.process_options(
             node,
             {
-                'sources': {'default': []},
-                'targets': {'default': []},
                 'type':    {'default': ''},
             },
             docname=env.docname
         )
+        self.process_options(
+            node,
+            {
+                'sources': {'default': []},
+                'targets': {'default': []},
+                'source':  {'default': '', 'is_pattern': True},
+                'target':  {'default': '', 'is_pattern': True},
+            },
+        )
+        for mutually_exclusive_options in ({'sources', 'source'}, {'targets', 'target'}):
+            option_amount = len(mutually_exclusive_options.intersection(self.options))
+            if option_amount != 1:
+                report_warning(f"item-link: expected exactly one of the following options but got {option_amount}: "
+                               f"{mutually_exclusive_options}", env.docname, self.lineno)
+                process_options_success = False
         if not process_options_success:
             return []
-
-        # Processing of the item-link items. They get added as additional relationships
-        # to the existing items. Should be done before converting anything to docutils.
-        for source in node['sources']:
-            for target in node['targets']:
-                try:
-                    env.traceability_collection.add_relation(source, node['type'], target)
-                except TraceabilityException as err:
-                    report_warning(err, env.docname, self.lineno)
-
-        # The ItemLink node has no final representation, so is removed from the tree
+        env.traceability_collection.add_intermediate_node(node)
         return [node]
