@@ -1,16 +1,18 @@
 import re
-from collections import OrderedDict, namedtuple
+from collections import namedtuple
 from copy import copy, deepcopy
 
 from docutils import nodes
 from docutils.parsers.rst import directives
-from natsort import natsorted
+from natsort import natsorted, natsort_keygen
 
 from mlx.traceability_exception import TraceabilityException, report_warning
 from mlx.traceable_base_directive import TraceableBaseDirective
 from mlx.traceable_base_node import TraceableBaseNode
 from mlx.traceable_item import TraceableItem
 
+natsort_key = natsort_keygen(key=lambda item: getattr(item, 'identifier', ''))
+
 
 def group_choice(argument):
     """Conversion function for the "group" option."""
@@ -20,6 +22,8 @@ def group_choice(argument):
 class ItemMatrix(TraceableBaseNode):
     '''Matrix for cross referencing documentation items'''
 
+    COVERAGE_REGEX = re.compile(r'([><]=?|==|!=)\s*[\d\./]+')
+
     def perform_replacement(self, app, collection):
         """
         Creates table with related items, printing their target references. Only source and target items matching
@@ -30,10 +34,15 @@ class ItemMatrix(TraceableBaseNode):
             collection (TraceableCollection): Collection for which to generate the nodes.
         """
         Rows = namedtuple('Rows', "sorted covered uncovered counters")
-        source_ids = collection.get_items(self['source'], attributes=self['filter-attributes'])
+        filters = {x: None for x in ['source', 'target']}
+        if self['filtertarget']:
+            filters['target'] = self['filter-attributes']
+        else:
+            filters['source'] = self['filter-attributes']
+        source_ids = collection.get_items(self['source'], attributes=filters['source'])
         targets_with_ids = []
         for target_regex in self['target']:
-            targets_with_ids.append(collection.get_items(target_regex))
+            targets_with_ids.append(collection.get_items(target_regex, attributes=filters['target']))
         top_node = self.create_top_node(self['title'], hide_title=self['hidetitle'])
         table = nodes.table()
         if self.get('classes'):
@@ -80,7 +89,8 @@ class ItemMatrix(TraceableBaseNode):
             if ' | ' in self['type']:
                 mapping_via_intermediate = self.linking_via_intermediate(source_ids, targets_with_ids, collection)
 
-        duplicate_source_count = 0
+        duplicate_count_total = 0
+        duplicate_count_covered = 0
         rows = Rows([], [], [], [0, 0])
         for source_id in source_ids:
             source_item = collection.get_item(source_id)
@@ -89,17 +99,35 @@ class ItemMatrix(TraceableBaseNode):
             covered = False
             rights = [[] for _ in range(int(bool(self['intermediate'])) + len(self['target']))]
             if mapping_via_intermediate:
-                covered = source_id in mapping_via_intermediate
-                if covered:
-                    args = [rows, source_item, rights, app]
-                    duplicate_source_count += self._store_source_via_intermediate(mapping_via_intermediate[source_id],
-                                                                                  *args)
+                intermediates = mapping_via_intermediate[source_id]
+                if not intermediates:
+                    self._store_data(rows, source_item, rights, False, app)
+                    continue
+                covered_intermediates = {intermediate: any(target_set for target_set in target_sets)
+                                         for intermediate, target_sets in intermediates.items()}
+
+                if self['coveredintermediates']:
+                    covered = all(covered_intermediates.values())
+                else:
+                    covered = any(covered_intermediates.values())
+
+                if self['splitintermediates']:
+                    for intermediate, target_sets in intermediates.items():
+                        self._store_row_with_intermediate({intermediate: target_sets},
+                                                          rows, source_item, rights, covered, app)
+                    duplicate_count_for_source = len(intermediates) - 1
+                    duplicate_count_total += duplicate_count_for_source
+                    duplicate_count_covered += duplicate_count_for_source if covered else 0
+                else:
+                    self._store_row_with_intermediate({intermediate: target_sets for intermediate, target_sets
+                                                       in intermediates.items()
+                                                       if covered and covered_intermediates[intermediate]},
+                                                      rows, source_item, rights, covered, app)
             else:
                 has_external_target = self.add_external_targets(rights, source_item, external_relationships, app)
                 has_internal_target = self.add_internal_targets(rights, source_id, targets_with_ids, relationships,
                                                                 collection)
                 covered = has_external_target or has_internal_target
-            if not (covered and mapping_via_intermediate):
                 self._store_data(rows, source_item, rights, covered, app)
 
         if not source_ids:
@@ -107,7 +135,7 @@ class ItemMatrix(TraceableBaseNode):
             for ext_rel in external_relationships:
                 external_targets = collection.get_external_targets(self['source'], ext_rel)
                 # natural sorting on source
-                for ext_source, target_ids in OrderedDict(natsorted(external_targets.items())).items():
+                for ext_source, target_ids in natsorted(external_targets.items()):
                     covered = False
                     source_link = self.make_external_item_ref(app, ext_source, ext_rel)
                     rights = [[] for _ in range(len(self['target']))]
@@ -117,8 +145,8 @@ class ItemMatrix(TraceableBaseNode):
 
         tgroup += self._build_table_body(rows, self['group'], self['onlycovered'], self['onlyuncovered'])
 
-        count_total = rows.counters[0] + rows.counters[1] - duplicate_source_count
-        count_covered = rows.counters[0] - duplicate_source_count
+        count_total = rows.counters[0] + rows.counters[1] - duplicate_count_total
+        count_covered = rows.counters[0] - duplicate_count_covered
         try:
             percentage = 100 * count_covered / count_total
         except ZeroDivisionError:
@@ -138,7 +166,8 @@ class ItemMatrix(TraceableBaseNode):
             p_node += txt
             top_node += p_node
 
-        top_node += table
+        if number_of_columns:
+            top_node += table
         self.replace_self(top_node)
 
     def _build_table_body(self, rows, group, onlycovered, onlyuncovered):
@@ -240,18 +269,18 @@ class ItemMatrix(TraceableBaseNode):
             right_cells (list): List of empty lists to fill with intermediates items followed by target items
             linked_items (dict): Mapping of intermediate items to the list of sets of target items per target
         """
-        for intermediate_item in linked_items:
-            right_cells[0].append(intermediate_item)
-
         # avoid duplicate target IDs in the same cell due to multiple intermediates with the same target item
         added_items_per_column = {}
-        for targets in linked_items.values():
-            for idx, target_items in enumerate(targets):
+        for intermediate_item, targets in linked_items.items():
+            right_cells[0].append(intermediate_item)
+            for idx, target_items in enumerate(targets, start=1):
                 if idx not in added_items_per_column:
                     added_items_per_column[idx] = set()
                 for target_item in target_items.difference(added_items_per_column[idx]):
-                    right_cells[idx + 1].append(target_item)
+                    right_cells[idx].append(target_item)
                     added_items_per_column[idx].add(target_item)
+                right_cells[idx].sort(key=natsort_key)
+        right_cells[0].sort(key=natsort_key)
 
     def add_external_targets(self, right_cells, source_item, external_relationships, app):
         """ Adds links to external targets for given source to the list of data per column
@@ -267,7 +296,7 @@ class ItemMatrix(TraceableBaseNode):
         """
         has_external_target = False
         for external_relationship in external_relationships:
-            for target_id in source_item.iter_targets(external_relationship):
+            for target_id in source_item.yield_targets_sorted(external_relationship):
                 ext_item_ref = self.make_external_item_ref(app, target_id, external_relationship)
                 for cell in right_cells:
                     cell.append(ext_item_ref)
@@ -299,8 +328,6 @@ class ItemMatrix(TraceableBaseNode):
     def linking_via_intermediate(self, source_ids, targets_with_ids, collection):
         """ Maps source IDs to IDs of target items that are linked via an itermediate item per target
 
-        Only covered source IDs are stored.
-
         Args:
             source_ids (list): List of item IDs of source items
             targets_with_ids (list): List of lists, which contain target IDs to take into consideration, per target
@@ -321,42 +348,26 @@ class ItemMatrix(TraceableBaseNode):
         for idx, rel in enumerate(links_with_relationships[0]):
             links_with_relationships[0][idx] = collection.get_reverse_relation(rel)
 
-        source_to_links_map = {}
-        excluded_source_ids = set()
+        source_to_links_map = {source_id: {} for source_id in source_ids}
         for intermediate_id in collection.get_items(self['intermediate'], sort=bool(self['intermediatetitle'])):
             intermediate_item = collection.get_item(intermediate_id)
 
             potential_source_ids = set()
             for reverse_rel in links_with_relationships[0]:
-                potential_source_ids.update(intermediate_item.iter_targets(reverse_rel, sort=False))
+                potential_source_ids.update(intermediate_item.yield_targets(reverse_rel))
             # apply :source: filter
             potential_source_ids = potential_source_ids.intersection(source_ids)
-            potential_source_ids = potential_source_ids.difference(excluded_source_ids)
-            if not potential_source_ids:
-                continue
 
             potential_target_ids = set()
             for forward_rel in links_with_relationships[1]:
-                potential_target_ids.update(intermediate_item.iter_targets(forward_rel, sort=False))
-            if not potential_target_ids:
-                if self['coveredintermediates']:
-                    excluded_source_ids.update(potential_source_ids)
-                continue
+                potential_target_ids.update(intermediate_item.yield_targets(forward_rel))
             # apply :target: filter
-            covered = False
             actual_targets = []
             for target_ids in targets_with_ids:
                 linked_target_ids = potential_target_ids.intersection(target_ids)
-                if linked_target_ids:
-                    covered = True
                 actual_targets.append(set(collection.get_item(id_) for id_ in linked_target_ids))
 
-            if covered:
-                self._store_targets(source_to_links_map, potential_source_ids, actual_targets, intermediate_item)
-            elif self['coveredintermediates']:
-                excluded_source_ids.update(potential_source_ids)
-        for source_id in excluded_source_ids:
-            source_to_links_map.pop(source_id, None)
+            self._store_targets(source_to_links_map, potential_source_ids, actual_targets, intermediate_item)
         return source_to_links_map
 
     @staticmethod
@@ -376,25 +387,7 @@ class ItemMatrix(TraceableBaseNode):
                 source_to_links_map[source_id] = {}
             source_to_links_map[source_id][intermediate_item] = targets
 
-    def _store_source_via_intermediate(self, linked_items, *args):
-        """ Stores row(s) for a source, linking targets via intermediates
-
-        Args:
-            linked_items (dict): Mapping of all intermediate IDs to the list of sets of target items per target
-
-        Returns:
-            int: Number of rows that have been added with a duplicate source ID
-        """
-        duplicate_source_count = 0
-        if self['splitintermediates']:
-            for intermediate, targets in linked_items.items():
-                self._store_row_with_intermediate({intermediate: targets}, *args)
-            duplicate_source_count += len(linked_items) - 1
-        else:
-            self._store_row_with_intermediate(linked_items, *args)
-        return duplicate_source_count
-
-    def _store_row_with_intermediate(self, linked_items, rows, source, empty_right_cells, app):
+    def _store_row_with_intermediate(self, linked_items, rows, source, empty_right_cells, *args):
         """ Stores a row for a source, linking targets via one or all intermediates
 
         Args:
@@ -402,11 +395,10 @@ class ItemMatrix(TraceableBaseNode):
             rows (Rows): Rows namedtuple object to extend
             source (TraceableItem): Source item
             empty_right_cells (list): List of empty lists to fill with intermediates items, followed by target items
-            app (sphinx.application.Sphinx): Sphinx application object
         """
         right_cells = deepcopy(empty_right_cells)
         self.add_all_targets(right_cells, linked_items)
-        self._store_data(rows, source, right_cells, True, app)
+        self._store_data(rows, source, right_cells, *args)
 
     def _store_data(self, rows, source, right_cells, covered, app):
         """ Stores the data in one or more rows in the given Rows object.
@@ -452,6 +444,7 @@ class ItemMatrix(TraceableBaseNode):
                     items = target_items
                 elif row_idx < len(target_items):
                     items = [target_items[row_idx]]
+                items.sort(key=natsort_key)
                 row += self._create_cell_for_items(items, app)
             # target columns: attributes and extra relations
             target_attribute_cells = []
@@ -489,7 +482,7 @@ class ItemMatrix(TraceableBaseNode):
         covered = False
         for idx, target_regex in enumerate(self['target']):
             for target in target_items:
-                if re.match(target_regex, target.get_id()):
+                if target_regex and target_regex.match(target.identifier):
                     target_cells[idx].append(target)
                     covered = True
         return covered
@@ -509,7 +502,7 @@ class ItemMatrix(TraceableBaseNode):
             if isinstance(entry, nodes.Node):
                 cell += entry
             else:
-                cell += self.make_internal_item_ref(app, entry.get_id())
+                cell += self.make_internal_item_ref(app, entry.identifier)
         return cell
 
     def _create_cells_for_info_cols(self, item, values, app):
@@ -533,7 +526,7 @@ class ItemMatrix(TraceableBaseNode):
 
     @staticmethod
     def _create_cell_for_attribute(item, attribute):
-        """ Creates a cell with the item's attribute value the given attribute.
+        """ Creates a cell with the item's attribute value for the given attribute.
 
         Args:
             item (TraceableItem): TraceableItem instance
@@ -561,7 +554,7 @@ class ItemMatrix(TraceableBaseNode):
         """
         cell = nodes.entry('')
         if not isinstance(item, nodes.paragraph):
-            for linked_item in item.iter_targets(relation):
+            for linked_item in item.yield_targets_sorted(relation):
                 if self.is_relation_external(relation):
                     cell += self.make_external_item_ref(app, linked_item, relation)
                 else:
@@ -578,7 +571,7 @@ class ItemMatrix(TraceableBaseNode):
         """
         if self['coverage']:
             pattern = r'([><]=?|==|!=)\s*[\d\./]+'
-            if re.fullmatch(pattern, self['coverage']):
+            if self.COVERAGE_REGEX.fullmatch(self['coverage']):
                 expression = '{} {}'.format(percentage, self['coverage'])
                 if not eval(expression):  # pylint: disable=eval-used
                     report_warning('Item-matrix with title {!r} has bad coverage: {} evaluates to False'
@@ -600,7 +593,7 @@ class ItemMatrixDirective(TraceableBaseDirective):
     Syntax::
 
       .. item-matrix:: title
-         :target: regexp
+         :target: regexp ...
          :source: regexp
          :intermediate: regexp
          :<<attribute>>: regexp
@@ -623,6 +616,7 @@ class ItemMatrixDirective(TraceableBaseDirective):
          :nocaptions:
          :onlycaptions:
          :hidetitle:
+         :filtertarget:
     """
     # Optional argument: title (whitespace allowed)
     optional_arguments = 1
@@ -638,9 +632,7 @@ class ItemMatrixDirective(TraceableBaseDirective):
         'type': directives.unchanged,  # relationship types separated by space
         'sourcetype': directives.unchanged,  # relationship types separated by space
         'sourcecolumns': directives.unchanged,  # attributes separated by space
-        'sourceattributes': directives.unchanged,  # deprecated in 9.x
         'targetcolumns': directives.unchanged,  # attributes separated by space
-        'targetattributes': directives.unchanged,  # deprecated in 9.x
         'hidesource': directives.flag,
         'hidetarget': directives.flag,
         'splitintermediates': directives.flag,
@@ -654,6 +646,7 @@ class ItemMatrixDirective(TraceableBaseDirective):
         'nocaptions': directives.flag,
         'onlycaptions': directives.flag,
         'hidetitle': directives.flag,
+        'filtertarget': directives.flag,
     }
     # Content disallowed
     has_content = False
@@ -671,14 +664,14 @@ class ItemMatrixDirective(TraceableBaseDirective):
 
         self.process_title(node, 'Traceability matrix of items')
 
-        self.add_found_attributes(node)
+        self.add_found_attributes(node, is_pattern=True)
 
         self.process_options(
             node,
             {
-                'target':            {'default': ['']},
-                'intermediate':      {'default': ''},
-                'source':            {'default': ''},
+                'target':            {'default': [''], 'is_pattern': True},
+                'intermediate':      {'default': '', 'is_pattern': True},
+                'source':            {'default': '', 'is_pattern': True},
                 'targettitle':       {'default': ['Target'], 'delimiter': ','},
                 'sourcetitle':       {'default': 'Source'},
                 'intermediatetitle': {'default': ''},
@@ -705,22 +698,13 @@ class ItemMatrixDirective(TraceableBaseDirective):
         if number_of_targets != number_of_targettitles:
             raise TraceabilityException(
                 "Item-matrix directive should have the same number of values for the options 'target' and "
-                "'targettitle'. Got target: {targets} and targettitle: {titles}"
-                .format(targets=node['target'], titles=node['targettitle']),
+                "'targettitle'. Got target: {targets!r} and targettitle: {titles!r}"
+                .format(targets=self.options['target'], titles=self.options['targettitle']),
                 docname=env.docname)
 
         if node['type']:
             self.check_relationships(node['type'].replace(' | ', ' ').split(' '), env)
         self.check_relationships(node['sourcetype'], env)
-
-        if 'sourceattributes' in self.options:
-            report_warning("Option 'sourceattributes' will be deprecated in version 9.x in favor of 'sourcecolumns'",
-                           docname=env.docname, lineno=self.lineno)
-            self.options['sourcecolumns'] = self.options['sourceattributes']
-        if 'targetattributes' in self.options:
-            report_warning("Option 'targetattributes' will be deprecated in version 9.x in favor of 'targetcolumns'",
-                           docname=env.docname, lineno=self.lineno)
-            self.options['targetcolumns'] = self.options['targetattributes']
         self.add_attributes_and_relations(node, 'sourcecolumns', app.env.traceability_collection.relations)
         self.add_attributes_and_relations(node, 'targetcolumns', app.env.traceability_collection.relations)
         if number_of_targets > 1 and node['targetcolumns']:
@@ -739,6 +723,7 @@ class ItemMatrixDirective(TraceableBaseDirective):
         self.check_option_presence(node, 'coveredintermediates')
         self.check_option_presence(node, 'stats')
         self.check_option_presence(node, 'hidetitle')
+        self.check_option_presence(node, 'filtertarget')
 
         if node['onlycovered'] and node['onlyuncovered']:
             raise TraceabilityException(
