@@ -6,8 +6,8 @@ Traceability plugin
 Sphinx extension for reStructuredText that added traceable documentation items.
 See readme for more details.
 """
-
-from collections import OrderedDict, namedtuple
+import warnings
+from collections import namedtuple
 from re import fullmatch, match
 from os import path
 
@@ -36,6 +36,7 @@ from mlx.directives.item_link_directive import ItemLink, ItemLinkDirective
 from mlx.directives.item_list_directive import ItemList, ItemListDirective
 from mlx.directives.item_matrix_directive import ItemMatrix, ItemMatrixDirective
 from mlx.directives.item_pie_chart_directive import ItemPieChart, ItemPieChartDirective
+from mlx.directives.item_relink_directive import ItemRelink, ItemRelinkDirective
 from mlx.directives.item_tree_directive import ItemTree, ItemTreeDirective
 
 ItemInfo = namedtuple('ItemInfo', 'attr_val mr_id')
@@ -46,7 +47,7 @@ def generate_color_css(app, hyperlink_colors):
 
     Args:
         app: Sphinx application object to use.
-        hyperlink_colors (OrderedDict): Ordered dict with regex strings as keys and list/tuple of strings as values.
+        hyperlink_colors (dict): Dictionary with regex strings as keys and list/tuple of strings as values.
     """
     class_names = app.config.traceability_class_names
     with open(path.join(path.dirname(__file__), 'assets', 'hyperlink_colors.css'), 'w') as css_file:
@@ -190,25 +191,23 @@ class PendingItemXref(TraceableBaseNode):
 
 # -----------------------------------------------------------------------------
 # Event handlers
+def perform_consistency_check(app, env):
+    """Called once in between Sphinx' read stage and write stage.
 
-def perform_consistency_check(app, doctree):
-    """
-    New in sphinx 1.6: consistency checker callback
-
-    Used to perform the self-test on the collection of items
+    Used to perform (re)linking of item-link and item-relink and perform the self-test on the collection of items.
 
     If the ``checklist_item_regex`` is configured, a warning is reported
     for each item ID that matches it and is not defined as a checklist-item.
     """
-    env = app.builder.env
-
+    env.traceability_collection.process_intermediate_nodes()
+    ItemRelink.remove_placeholders(env.traceability_collection)
     try:
         env.traceability_collection.self_test(app.config.traceability_notifications.get('undefined-reference'))
     except TraceabilityException as err:
-        report_warning(str(err), err.get_document())
-    except MultipleTraceabilityExceptions as errs:
-        for err in errs.iter():
-            report_warning(str(err), err.get_document())
+        report_warning(str(err), err.docname)
+    except MultipleTraceabilityExceptions as errors:
+        for err in errors:
+            report_warning(str(err), err.docname)
 
     if app.config.traceability_json_export_path:
         fname = app.config.traceability_json_export_path
@@ -217,6 +216,9 @@ def perform_consistency_check(app, doctree):
     if app.config.traceability_hyperlink_colors:
         app.add_css_file('hyperlink_colors.css')
         generate_color_css(app, app.config.traceability_hyperlink_colors)
+        if [regex for regex in app.config.traceability_hyperlink_colors if not regex.startswith('^')]:
+            warnings.warn('Regexes in traceability_hyperlink_colors will be handled by re.match instead of re.search '
+                          'in mlx.traceability>=9', DeprecationWarning)
 
     regex = app.config.traceability_checklist.get('checklist_item_regex')
     if regex is not None and app.config.traceability_checklist['has_checklist_items']:
@@ -234,7 +236,9 @@ def process_item_nodes(app, doctree, fromdocname):
     node_classes = (
         AttributeLink,
         AttributeSort,
-        ItemLink, ItemMatrix,
+        ItemLink,
+        ItemRelink,
+        ItemMatrix,
         ItemPieChart,
         ItemAttributesMatrix,
         Item2DMatrix,
@@ -314,6 +318,7 @@ def initialize_environment(app):
     # generates placeholders when parsing the reverse relationships, the
     # database of items needs to be empty on every re-build.
     env.traceability_collection = TraceableCollection()
+    env.traceability_ref_nodes = {}
 
     app.config.traceability_checklist['has_checklist_items'] = False
     add_checklist_attribute(app.config.traceability_checklist,
@@ -374,7 +379,7 @@ def define_attribute(attr, app):
     """ Defines a new attribute. """
     attrobject = TraceableAttribute(attr, app.config.traceability_attributes[attr])
     if attr in app.config.traceability_attribute_to_string:
-        attrobject.set_name(app.config.traceability_attribute_to_string[attr])
+        attrobject.name = app.config.traceability_attribute_to_string[attr]
     else:
         report_warning('Traceability: attribute {attr} cannot be translated to string'.format(attr=attr))
     TraceableItem.define_attribute(attrobject)
@@ -383,7 +388,8 @@ def define_attribute(attr, app):
 def query_checklist(settings, attr_values):
     """ Queries specified API host name for the description of the specified merge request.
 
-    Reports a warning if the API host name is invalid or the response does not contain a description.
+    Reports a warning if the API host name is invalid, something went wrong with the GET request of the PR/MR,
+    or the response does not contain a description.
 
     Args:
         settings (dict): Dictionary with the environment variables specified for the checklist feature.
@@ -392,38 +398,41 @@ def query_checklist(settings, attr_values):
     Returns:
         (dict) The query results with zero or more key-value pairs in the form of {item ID: ItemInfo}.
     """
-    query_results = {}
     headers = {}
-    if not settings.get('private_token'):
-        settings['private_token'] = ''
-    if 'github' in settings['api_host_name']:
+    private_token = settings.get('private_token', '')
+    api_host_name = settings['api_host_name'].rstrip('/')
+    git_platform = settings.get('git_platform', api_host_name).lower()
+    if 'github' in git_platform:
         # explicitly request the v3 version of the REST API
         headers['Accept'] = 'application/vnd.github.v3+json'
-        if settings['private_token']:
-            headers['Authorization'] = 'token {}'.format(settings['private_token'])
-        base_url = "{}/repos/{}/pulls/".format(settings['api_host_name'].rstrip('/'),
-                                               settings['project_id'],)
+        if private_token:
+            headers['Authorization'] = 'token {}'.format(private_token)
+        base_url = "{}/repos/{}/pulls/".format(api_host_name, settings['project_id'])
         key = 'body'
-    elif 'gitlab' in settings['api_host_name']:
-        headers['PRIVATE-TOKEN'] = settings['private_token']
-        base_url = "{}/projects/{}/merge_requests/".format(settings['api_host_name'].rstrip('/'),
-                                                           settings['project_id'],)
+    elif 'gitlab' in git_platform:
+        headers['PRIVATE-TOKEN'] = private_token
+        base_url = "{}/projects/{}/merge_requests/".format(api_host_name, settings['project_id'])
         key = 'description'
     else:
+        report_warning("Failed to determine which GIT platform to use (GitHub or GitLab); "
+                       "please configure traceability_checklist['git_platform']")
         return {}
-
-    for merge_request_id in str(settings['merge_request_id']).split(','):
-        url = base_url + merge_request_id.strip()
-        with Session() as session:
-            with session.get(url, headers=headers) as response:
-                response = response.json()
-
-        description = response.get(key)
-        if description:
-            query_results = {**query_results, **_parse_description(description, attr_values, merge_request_id,
-                                                                   settings['checklist_item_regex'])}
-        else:
-            report_warning("The query did not return a description. URL = {}. Response = {}.".format(url, response))
+    query_results = {}
+    with Session() as session:
+        for merge_request_id in str(settings['merge_request_id']).split(','):
+            url = base_url + merge_request_id.strip()
+            try:
+                with session.get(url, headers=headers) as response:
+                    response = response.json()
+            except Exception as err:
+                report_warning("traceability_checklist: failed to GET {!r}: {!r}".format(url, err))
+                continue
+            description = response.get(key)
+            if description:
+                query_results = {**query_results, **_parse_description(description, attr_values, merge_request_id,
+                                                                       settings['checklist_item_regex'])}
+            else:
+                report_warning("The query did not return a description. URL = {}. Response = {}.".format(url, response))
     return query_results
 
 
@@ -460,9 +469,7 @@ def _parse_description(description, attr_values, merge_request_id, regex):
 # Extension setup
 def setup(app):
     """Extension setup"""
-
     # Javascript and stylesheet for the tree-view
-    # app.add_js_file('jquery.js') #note: can only be included once
     app.add_js_file('https://cdn.rawgit.com/aexmachina/jquery-bonsai/master/jquery.bonsai.js')
     app.add_css_file('https://cdn.rawgit.com/aexmachina/jquery-bonsai/master/jquery.bonsai.css')
     app.add_js_file('traceability.js')
@@ -470,9 +477,12 @@ def setup(app):
     # Configuration for exporting collection to json
     app.add_config_value('traceability_json_export_path', None, 'env')
 
-    # Configuration for adapting items through a callback
+    # Configuration for adapting items through a callback while processing the ``item`` directives
     app.add_config_value('traceability_callback_per_item', None, 'env')
 
+    # Configuration for inspecting items through a callback after all directives have been processed
+    app.add_config_value('traceability_inspect_item', None, 'env')
+
     # Create default attributes dictionary. Can be customized in conf.py
     app.add_config_value(
         'traceability_attributes',
@@ -481,7 +491,7 @@ def setup(app):
             'asil': '^(QM|[ABCD])$',
             'aspice': '^[123]$',
             'status': '^.*$',
-            'result': '(?i)^(pass|fail|error)$',
+            'result': '(?i)^(pass|fail|error|skip)$',
             'attendees': '^([A-Z]{3}[, ]*)+$',
             'assignee': '^.*$',
             'effort': r'^([\d\.]+(mo|[wdhm]) ?)+$',
@@ -577,7 +587,7 @@ def setup(app):
     app.add_config_value('traceability_tree_no_captions', False, 'env')
 
     # Configuration for customizing the color of hyperlinked items
-    app.add_config_value('traceability_hyperlink_colors', OrderedDict([]), 'env')
+    app.add_config_value('traceability_hyperlink_colors', {}, 'env')
     # Dictionary used by plugin to pass class names via application object
     app.add_config_value('traceability_class_names', {}, 'env')
 
@@ -608,6 +618,7 @@ def setup(app):
     app.add_directive('item-2d-matrix', Item2DMatrixDirective)
     app.add_directive('item-tree', ItemTreeDirective)
     app.add_directive('item-link', ItemLinkDirective)
+    app.add_directive('item-relink', ItemRelinkDirective)
     app.add_directive('attribute-link', AttributeLinkDirective)
     app.add_directive('attribute-sort', AttributeSortDirective)
 
@@ -619,4 +630,8 @@ def setup(app):
                                   innernodeclass=nodes.emphasis,
                                   warn_dangling=True))
 
-    return {'version': version}
+    return {
+        'version': version,
+        'parallel_read_safe': False,
+        'parallel_write_safe': True,
+    }
