@@ -5,6 +5,7 @@ from abc import abstractmethod, ABC
 from docutils import nodes
 from sphinx.errors import NoUri
 from sphinx.builders.latex import LaTeXBuilder
+from sphinx.util.osutil import SEP
 
 from mlx.traceability_exception import report_warning
 from mlx.traceable_item import TraceableItem
@@ -15,26 +16,30 @@ EXTERNAL_LINK_FIELDNAME = 'field'
 class TraceableBaseNode(nodes.General, nodes.Element, ABC):
     """ Base class for all Traceability node classes. """
 
-    def create_top_node(self, title, app=None):
-        ''' Creates the top node for the Element node. An admonition object with given title is created and returns.
+    def create_top_node(self, title, app=None, hide_title=False):
+        ''' Creates the top node for the Element node.
+
+        When the title should be shown, an admonition object with given title is added.
 
         Args:
             title (str): Title or item ID of the top node
             app (sphinx.application.Sphinx): Optional application object, needed when item ID is given to create link
+            hide_title (bool): True to add the title in an admonition node; False to return an empty container node
 
         Returns:
-            Top level replacement node to which other nodes can be appended
+            nodes.container: Top level replacement node to which other nodes can be appended
         '''
         top_node = nodes.container()
-        admon_node = nodes.admonition()
-        admon_node['classes'].append('item')
-        title_node = nodes.title()
-        if app:
-            title_node += self.make_internal_item_ref(app, title).children[0]
-        else:
-            title_node += nodes.Text(title)
-        admon_node += title_node
-        top_node += admon_node
+        if not hide_title:
+            admon_node = nodes.admonition()
+            admon_node['classes'].append('item')
+            title_node = nodes.title()
+            if app:
+                title_node += self.make_internal_item_ref(app, title).children[0]
+            else:
+                title_node += nodes.Text(title)
+            admon_node += title_node
+            top_node += admon_node
         return top_node
 
     @abstractmethod
@@ -52,37 +57,48 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
         paragraph. Reference text adds also a caption if it exists.
         """
         env = app.builder.env
+        ref_nodes = env.traceability_ref_nodes
+        display_option = 'default'
+        for option in (name for name in ('onlycaptions', 'nocaptions') if self.get(name)):
+            display_option = option
+            break
         item_info = env.traceability_collection.get_item(item_id)
+        link_item = item_info
         notification_item = None
-
         p_node = nodes.paragraph()
 
         # Only create link when target item (or notification item) exists, warn otherwise (in html and terminal)
-        if item_info.is_placeholder():
+        if item_info.is_placeholder:
             notification_item_id = app.config.traceability_notifications.get('undefined-reference')
-            notification_item = app.env.traceability_collection.get_item(notification_item_id)
+            notification_item = env.traceability_collection.get_item(notification_item_id)
             if not notification_item:
                 self.has_warned_about_undefined(item_info)
                 txt = nodes.Text('%s not defined, broken link' % item_id)
                 p_node.append(txt)
                 return p_node
+            link_item = notification_item
+        try:
+            if self['document'] == link_item.docname and hasattr(app.builder, 'link_suffix'):
+                # include filename so that the returned node can be reused on every page in the same directory
+                relative_path = link_item.docname.split(SEP)[-1] + app.builder.link_suffix
+            else:
+                relative_path = app.builder.get_relative_uri(self['document'], link_item.docname)
+        except NoUri:
+            # ignore if no URI can be determined, e.g. for LaTeX output
+            relative_path = ''
 
-        display_text, text_on_hover_node = self._get_caption_info(item_info)
+        if item_id not in ref_nodes:
+            ref_nodes[item_id] = {}
+        if display_option not in ref_nodes[item_id]:
+            ref_nodes[item_id][display_option] = {}
+        if relative_path in ref_nodes[item_id][display_option]:
+            return ref_nodes[item_id][display_option][relative_path]  # cached paragraph node
 
+        display_text, text_on_hover_node = self._get_caption_info(item_info)  # display original item
         newnode = nodes.reference('', '')
         innernode = nodes.emphasis(display_text, display_text)
-        try:
-            if not notification_item:
-                newnode['refuri'] = app.builder.get_relative_uri(self['document'], item_info.docname)
-                newnode['refuri'] += '#' + item_id
-                newnode['refdocname'] = item_info.docname
-            else:
-                newnode['refuri'] = app.builder.get_relative_uri(self['document'], notification_item.docname)
-                newnode['refuri'] += '#' + notification_item_id
-                newnode['refdocname'] = notification_item.docname
-        except NoUri:
-            # ignore if no URI can be determined, e.g. for LaTeX output :(
-            pass
+        newnode['refuri'] = f"{relative_path}#{link_item.identifier}"
+        newnode['refdocname'] = link_item.docname
 
         # change text color if item_id matches a regex in traceability_hyperlink_colors
         colors = self._find_colors_for_class(app.config.traceability_hyperlink_colors, item_id)
@@ -95,6 +111,7 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
         newnode.append(innernode)
         p_node += newnode
 
+        ref_nodes[item_id][display_option][relative_path] = p_node
         return p_node
 
     @staticmethod
@@ -129,26 +146,7 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
         Returns:
             bool: True if relationship is an external one, False otherwise
         '''
-        return True if re.search('^ext_.*', relation) else False
-
-    def is_item_top_level(self, env, item_id):
-        '''
-        Checks if item with given item ID is a top level item.
-
-        True, if the item is a top level item:
-
-        - given relation does not exist for given item,
-        - or given relation exists, but targets don't match the 'top' regexp.
-
-        False, otherwise.
-        '''
-        item = env.traceability_collection.get_item(item_id)
-        for relation in self['top_relation_filter']:
-            tgts = item.iter_targets(relation, sort=False)
-            for tgt in tgts:
-                if re.match(self['top'], tgt):
-                    return False
-        return True
+        return relation.startswith('ext_')
 
     def make_attribute_ref(self, app, attr_id, value=''):
         """
@@ -161,20 +159,20 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
 
         if attr_id in TraceableItem.defined_attributes:
             attr_info = TraceableItem.defined_attributes[attr_id]
-            attr_name = attr_info.get_name()
+            attr_name = attr_info.name
             if attr_info.docname:
                 newnode = nodes.reference('', '')
                 innernode = nodes.emphasis(attr_name + value, attr_name + value)
                 newnode['refdocname'] = attr_info.docname
                 try:
                     newnode['refuri'] = app.builder.get_relative_uri(self['document'], attr_info.docname)
-                    newnode['refuri'] += '#' + attr_info.get_name()
+                    newnode['refuri'] += '#' + attr_info.identifier
                 except NoUri:
                     # ignore if no URI can be determined, e.g. for LaTeX output :(
                     pass
                 newnode.append(innernode)
             else:
-                newnode = nodes.Text('{attr}{value}'.format(attr=attr_info.get_name(), value=value))
+                newnode = nodes.Text('{attr}{value}'.format(attr=attr_name, value=value))
         else:
             newnode = nodes.Text('{attr}{value}'.format(attr=attr_id, value=value))
         p_node += newnode
@@ -188,7 +186,7 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
         The regexp of the first item in the ordered dictionary has the highest priority.
 
         Args:
-            hyperlink_colors (OrderedDict): Ordered dict with regex strings as keys and list/tuple of strings as values.
+            hyperlink_colors (dict): Dictionary with regex strings as keys and list/tuple of strings as values.
             item_id (str): A traceability item ID.
 
         Returns:
@@ -211,15 +209,15 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
             nodes.inline/None: Inline node containing the item's caption or ID to be shown on hover, or None to not
                 show anything on hover.
         """
-        display_text = item_info.id
+        display_text = item_info.identifier
         hidden_node = None
         if item_info and item_info.caption:
             if self.get('onlycaptions'):
                 display_text = item_info.caption
-                hidden_node = nodes.inline('', item_info.id)
+                hidden_node = nodes.inline('', item_info.identifier)
                 hidden_node['classes'].append('popup_caption')
             elif not self.get('nocaptions'):
-                display_text = '{0.id} : {0.caption}'.format(item_info)
+                display_text = '{0.identifier} : {0.caption}'.format(item_info)
             else:
                 hidden_node = nodes.inline('', item_info.caption)
                 hidden_node['classes'].append('popup_caption')
@@ -232,8 +230,8 @@ class TraceableBaseNode(nodes.General, nodes.Element, ABC):
         Args:
             item_info (TraceableItem): TraceableItem object.
         """
-        if item_info.is_placeholder():
-            report_warning("Traceability: cannot link to '%s', item is not defined" % item_info.get_id(),
+        if item_info.is_placeholder:
+            report_warning("Traceability: cannot link to '%s', item is not defined" % item_info.identifier,
                            self['document'], self['line'])
             return True
         return False
