@@ -3,6 +3,7 @@ Storage classes for collection of traceable items
 '''
 import json
 import re
+from operator import attrgetter
 from pathlib import Path
 
 from natsort import natsorted
@@ -23,6 +24,7 @@ class TraceableCollection:
         self.relations = {}
         self.items = {}
         self.relations_sorted = {}
+        self._intermediate_nodes = []
 
     def add_relation_pair(self, forward, reverse=NO_RELATION_STR):
         '''
@@ -69,18 +71,17 @@ class TraceableCollection:
         Args:
             item (TraceableItem): Traceable item to add
         '''
-        itemid = item.get_id()
         # If the item already exists ...
-        if itemid in self.items:
-            olditem = self.items[itemid]
+        if item.identifier in self.items:
+            olditem = self.items[item.identifier]
             # ... and it's not a placeholder, log an error
-            if not olditem.is_placeholder():
-                raise TraceabilityException('duplicating {itemid}'.format(itemid=itemid), item.get_document())
+            if not olditem.is_placeholder:
+                raise TraceabilityException('duplicating {itemid}'.format(itemid=item.identifier), item.docname)
             # ... otherwise, update the item with new content
             else:
                 item.update(olditem)
         # add it
-        self.items[item.get_id()] = item
+        self.items[item.identifier] = item
 
     def get_item(self, itemid):
         '''
@@ -89,11 +90,9 @@ class TraceableCollection:
         Args:
             itemid (str): Identification of traceable item to get
         Returns:
-            TraceableItem: Object for traceable item
+            TraceableItem/None: Object for traceable item; None if the item was not found
         '''
-        if self.has_item(itemid):
-            return self.items[itemid]
-        return None
+        return self.items.get(itemid)
 
     def iter_items(self):
         '''
@@ -133,7 +132,7 @@ class TraceableCollection:
         source = self.items[source_id]
         # Error if relation is unknown
         if relation not in self.relations:
-            raise TraceabilityException('Relation {name} not known'.format(name=relation), source.get_document())
+            raise TraceabilityException('Relation {name} not known'.format(name=relation), source.docname)
         # Add forward relation
         source.add_target(relation, target_id)
         # When reverse relation exists, continue to create/adapt target-item
@@ -169,6 +168,13 @@ class TraceableCollection:
                 item.attribute_order = attributes
         return ignored_items
 
+    def add_intermediate_node(self, node):
+        self._intermediate_nodes.append(node)
+
+    def process_intermediate_nodes(self):
+        for node in sorted(self._intermediate_nodes, key=attrgetter('order')):
+            node.apply_effect(self)
+
     def export(self, fname):
         '''
         Exports collection content. The target location of the json file gets created if it doesn't exist yet.
@@ -201,10 +207,10 @@ class TraceableCollection:
         for itemid in self.items:
             item = self.get_item(itemid)
             # Only for relevant items, filtered on document name
-            if docname is not None and item.get_document() != docname and item.get_document() is not None:
+            if docname is not None and item.docname != docname and item.docname is not None:
                 continue
             # Check if docname of notification item will be used
-            if item.get_document() is None and notification_item:
+            if item.docname is None and notification_item:
                 continue
             # On item level
             try:
@@ -217,23 +223,30 @@ class TraceableCollection:
                 rev_relation = self.get_reverse_relation(relation)
                 if rev_relation == self.NO_RELATION_STR:
                     continue
-                for tgt in item.iter_targets(relation):
+                for tgt in item.yield_targets(relation):
                     # Target item exists?
                     if tgt not in self.items:
                         errors.append(TraceabilityException("{source} {relation} {target}, but {target} is not known"
                                                             .format(source=itemid,
                                                                     relation=relation,
                                                                     target=tgt),
-                                                            item.get_document()))
+                                                            item.docname))
                         continue
                     # Reverse relation exists?
                     target = self.get_item(tgt)
-                    if itemid not in target.iter_targets(rev_relation, sort=False):
+                    if itemid not in target.yield_targets(rev_relation):
                         errors.append(TraceabilityException("No automatic reverse relation: {source} {relation} "
                                                             "{target}".format(source=tgt,
                                                                               relation=rev_relation,
                                                                               target=itemid),
-                                                            item.get_document()))
+                                                            item.docname))
+                    # Circular relation exists?
+                    for target_of_target in target.yield_targets(relation):
+                        if target_of_target in item.yield_targets(rev_relation):
+                            errors.append(TraceabilityException(
+                                "Circular relationship found: {} {} {} {} {} {} {}"
+                                .format(itemid, relation, tgt, relation, target_of_target, relation, itemid),
+                                item.docname))
         if errors:
             raise MultipleTraceabilityExceptions(errors)
 
@@ -265,62 +278,82 @@ class TraceableCollection:
         if source_id not in self.items:
             return False
         source = self.items[source_id]
-        if not source or source.is_placeholder():
+        if not source or source.is_placeholder:
             return False
         if target_id not in self.items:
             return False
         target = self.items[target_id]
-        if not target or target.is_placeholder():
+        if not target or target.is_placeholder:
             return False
         if not relations:
             relations = self.relations
         return self.items[source_id].is_related(relations, target_id)
 
-    def get_items(self, regex, attributes=None, sortattributes=None, reverse=False):
+    def get_items(self, regex, attributes=None, sortattributes=None, reverse=False, sort=True):
         '''
         Get all items that match a given regular expression
 
         Placeholders are excluded
 
         Args:
-            regex (str): Regex to match the items in this collection against
+            regex (str/re.Pattern): Regex pattern or object to match the items in this collection against
             attributes (dict): Dictionary with attribute-regex pairs to match the items in this collection against
             sortattributes (list): List of attributes on which to alphabetically sort the items
             reverse (bool): True for reverse sorting
+            sort (bool): When sortattributes is falsy: True to enable natural sorting, False to disable sorting
+
         Returns:
             list: A sorted list of item-id's matching the given regex. Sorting is done naturally when sortattributes is
             unused.
         '''
         matches = []
-        for itemid in self.items:
-            if self.items[itemid].is_placeholder():
+        for itemid, item in self.items.items():
+            if item.is_placeholder:
                 continue
-            if self.items[itemid].is_match(regex) and \
-                    (not attributes or self.items[itemid].attributes_match(attributes)):
+            if item.is_match(regex) and (not attributes or item.attributes_match(attributes)):
                 matches.append(itemid)
         if sortattributes:
-            matches = sorted(matches, key=lambda itemid: self.get_item(itemid).get_attributes(sortattributes),
-                             reverse=reverse)
-        else:
-            matches = natsorted(matches, reverse=reverse)
+            return sorted(matches, key=lambda itemid: self.get_item(itemid).get_attributes(sortattributes),
+                          reverse=reverse)
+        elif sort:
+            return natsorted(matches, reverse=reverse)
         return matches
 
+    def get_item_objects(self, regex, attributes=None):
+        ''' Get all items that match a given regular expression as TraceableItem instances.
+
+        Placeholders are excluded.
+
+        Args:
+            regex (str): Regex to match the items in this collection against
+            attributes (dict): Dictionary with attribute-regex pairs to match the items in this collection against
+
+        Returns:
+            generator: An iterable of items matching the given regex.
+        '''
+        for item in self.items.values():
+            if item.is_placeholder:
+                continue
+            if item.is_match(regex) and (not attributes or item.attributes_match(attributes)):
+                yield item
+
     def get_external_targets(self, regex, relation):
         ''' Get all external targets for a given external relation with the IDs of their linked internal items
 
         Args:
-            regex (str): Regex to match the external target
+            regex (str/re.Pattern): Regex pattern or object to match the external target
             relation (str): External relation
         Returns:
             dict: Dictionary mapping external targets to the IDs of their linked internal items
         '''
         external_targets_to_item_ids = {}
         for item_id, item in self.items.items():
-            for target in item.iter_targets(relation):
-                if not re.match(regex, target):
+            for target in item.yield_targets(relation):
+                try:
+                    match = regex.match(target)
+                except AttributeError:
+                    match = re.match(regex, target)
+                if not match:
                     continue
-                if target not in external_targets_to_item_ids:
-                    external_targets_to_item_ids[target] = [item_id]
-                else:
-                    external_targets_to_item_ids[target].append(item_id)
+                external_targets_to_item_ids.setdefault(target, []).append(item_id)
         return external_targets_to_item_ids
